#include<stdio.h>
#include<assert.h>
#include<malloc.h>
#include<memory.h>
#include<stdio.h>

//! \file memory2m2.cpp
//   - .      
//      .
//        void*    _
//      ,     _
// (c) Proteus   lawnmower-man@mail.ru


#define MEM_ALIGN	4			//!<    
#define PAGE_SIZE   1024		//!<    

#define LIST_LINK 0		//!<   
#define FREE_LINK 1		//!<   

typedef unsigned short word;
typedef unsigned char  byte;
typedef unsigned int   uint;
typedef word index_t;			//!<    

//! 
struct stat_st{
	int pages_count;		//!<  
	int free_pages_count;	//!<     
	int all_space;			//!<    
	int free_space;			//!<    
	int lost_space;			//!<    
	int obj_count;			//!<   
};



#pragma pack(push,1)
//!      
class StubAllocator{
	const uint item_size;	//!<   
	//!      
	struct bank_st{
		void *prev[2];		//!<    (  )
		void *next[2];		//!<    (  )
		index_t free_index;	//!<      
		index_t free_count;	//!<   
	};
	void *bank_list_;						//!<   
	void *free_list_;						//!<   
	void NewBank();							//!<    
	void *ListInsert(void*,void*,uint i);	//!<      
	void *ListRemove(void*,uint i);			//!<     
	//!      
	const uint PageSize()
	{
		return PAGE_SIZE;
	}
	//!      
	const uint PageCount()
	{
		return (PageSize()-sizeof(bank_st))/(item_size+sizeof(index_t));
	}
	//!   
	//! \param data   
	//! \param index     
	void *Elem(void *data,int index)
	{
		return (byte*)data+index*(item_size+sizeof(index_t));
	}
	//!   ,    
	//! \param data  
	index_t &ElemIndex(void *data)
	{
		return *(index_t*)((byte*)data+item_size);
	}
	//!   
	//! \param data   
	//! \param index     
	index_t &Index(void *data,int index)
	{
		return ElemIndex(Elem(data,index));
	}
	//!     
	//! \param data   
	bank_st *Bank(void *data)
	{
		return (bank_st*)((byte*)data+PageSize()-sizeof(bank_st));
	}
	//@]
public:
	StubAllocator(uint);
	~StubAllocator();
	void *New();				//!<  
	bool Free(void*);			//!<  
	void Clear();				//!<     
	void Stat(stat_st*);		//!<   
	inline bool isCorrect();	//!<    (   )
};


//!     
class Allocator{
	StubAllocator **array;	//!<   
	uint capablity;			//!<   
	void Resize(uint);		//!<    
public:
	Allocator();
	~Allocator();
	void *New(uint size);			//!<   
	void Free(void *p,uint size);	//!<  
	void Stat();					//!<   
};
#pragma pack(pop)


StubAllocator::StubAllocator(uint size):item_size(size)
{
	assert(isCorrect());
	bank_list_=free_list_=0;
}


StubAllocator::~StubAllocator()
{
	Clear();
}


//!  .       
void StubAllocator::Clear()
{
	free_list_=0;
	if(bank_list_) {
		bank_st *bank=Bank(bank_list_);
		bank->prev[LIST_LINK]=bank->next[LIST_LINK]=0;
	}
	while(bank_list_) {
		void *next=Bank(bank_list_)->next[LIST_LINK];
		delete bank_list_;
		bank_list_=next;
	}
}


//!     
//! \param list 
//! \param bank  
//! \param i        (   )
//! \result    (      )
void *StubAllocator::ListInsert(void *list_,void *bank_,unsigned int i)
{
	bank_st *bank=Bank(bank_);
	if(!list_) list_=bank->next[i]=bank->prev[i]=bank_;
	else {
		bank_st *list=Bank(list_);
		bank->prev[i]=list->prev[i];
		bank->next[i]=list_;
		list->prev[i]=bank_;
		Bank(bank->prev[i])->next[i]=bank_;
	}
	return list_;
}


//!     
//! \param list  (  )
//! \param i        (   )
//! \result   
void *StubAllocator::ListRemove(void *list_,unsigned i)
{
	void *prev_;
	bank_st *list=Bank(list_);
	if(!list_ || list->next[i]==list_) return 0;
	prev_=list->prev[i];
	list_=list->next[i];
	Bank(prev_)->next[i]=list_;
	Bank(list_)->prev[i]=prev_;
	return list_;
}


//!      
void StubAllocator::NewBank()
{
	unsigned int n;
	//  
	void *bank_=new byte[PageSize()];
	Bank(bank_)->free_count=PageCount();
	Bank(bank_)->free_index=0;
	//    
	for(n=0; n<PageCount()-1; n++) Index(bank_,n)=n+1; Index(bank_,n)=(index_t)-1;
	//     (     )
	bank_list_=ListInsert(bank_list_,bank_,LIST_LINK);
	free_list_=ListInsert(free_list_,bank_,FREE_LINK);
}


//!   (     )
//! \result true    
bool StubAllocator::Free(void *obj)
{
	int index=*(index_t*)((byte*)obj+item_size);
	//   
	void *bank_=Elem(obj,-index);
	bank_st *bank=Bank(bank_);
	//     ,      
	if(!bank->free_count) free_list_=ListInsert(free_list_,bank_,FREE_LINK);
	//      
	ElemIndex(obj)=bank->free_index;
	bank->free_index=index;
	bank->free_count++;
	//            (     )
	if(bank->free_count<Bank(free_list_)->free_count) free_list_=bank_;
	//      ,   )
	if(bank->free_count==(index_t)PageCount()) {
		bank_list_=ListRemove(bank_,LIST_LINK);
		void *temp_=ListRemove(bank_,FREE_LINK);
		if(bank_==free_list_) free_list_=temp_; //     if(!temp) free_list=0,     
		delete bank_;
	}
	return !bank_list_;
}


//!   
void* StubAllocator::New()
{
	//    ,  
	if(!free_list_) NewBank();
	//     
	uint index=Bank(free_list_)->free_index;
	void *item_=Elem(free_list_,index);
	bank_st *free_list=Bank(free_list_);
	free_list->free_index=ElemIndex(item_);
	free_list->free_count--;
	ElemIndex(item_)=index;
	//      ,      
	if(!free_list->free_count) free_list_=ListRemove(free_list_,FREE_LINK);
	// ..
	return item_;
}

//!    (     )
//! \result false    
inline bool StubAllocator::isCorrect()
{
	return (index_t)-1>PageCount();
}


void StubAllocator::Stat(stat_st *stat)
{
#ifdef _DEBUG
	int t;
	void *bank_=bank_list_;
	if(bank_) do {
		t=PageCount()*(item_size+sizeof(index_t));
		stat->pages_count++;
		stat->all_space+=t;
		stat->lost_space+=PageSize()-sizeof(bank_st)-t;
		bank_=Bank(bank_)->next[LIST_LINK];
	}while(bank_!=bank_list_);
	bank_=free_list_;
	if(bank_) do {
		stat->free_pages_count++;
		stat->free_space+=Bank(bank_)->free_count*(item_size+sizeof(index_t));
		bank_=Bank(bank_)->next[FREE_LINK];
	}while(bank_!=free_list_);
#endif
}


Allocator::Allocator()
{
	array=0; capablity=0;
}


Allocator::~Allocator()
{
	for(uint n=0; n<capablity; n++) delete array[n];
	free(array);
}


//! \param size    (  ,   )
void Allocator::Resize(uint size)
{
	if(size<capablity) return;
	uint up=capablity;
	while(up<=size) up+=1+up/4;
	array=(StubAllocator**)realloc(array,up*sizeof(StubAllocator*));
	while(capablity<up) array[capablity++]=0;
}


//! \param size   
//! \result   
void *Allocator::New(uint size)
{
	if(size>PAGE_SIZE/5) return malloc(size);
	int n=(size+MEM_ALIGN-1+sizeof(index_t))/MEM_ALIGN;
	size=n*MEM_ALIGN-sizeof(index_t);
	Resize(n);
	if(!array[n]) array[n]=new StubAllocator(size);
	return array[n]->New();
}


//! \param p  
//! \param size  
void Allocator::Free(void *p,uint size)
{
	if(size>PAGE_SIZE/5) free(p); else {
		int n=(size+MEM_ALIGN-1+sizeof(index_t))/MEM_ALIGN;
		if(array[n]->Free(p)) delete array[n],array[n]=0;
	}
}


void Allocator::Stat()
{
#ifdef _DEBUG
	uint n;
	stat_st stat; memset(&stat,0,sizeof(stat));
	for(n=0; n<capablity; n++) if(array[n]) array[n]->Stat(&stat),stat.obj_count++;
	printf("[stat]\n");
	printf("pages count     : %d\n",stat.pages_count);		//  
	printf("free pages count: %d\n",stat.free_pages_count);	//     
	printf("all space       : %d\n",stat.all_space);		//    
	printf("free space      : %d\n",stat.free_space);		//    
	printf("lost space      : %d\n",stat.lost_space);		//    
	printf("obj count       : %d\n",stat.obj_count);		//   
#endif
}



#include<stdlib.h>
#define COUNT 9000		//    
#define SIZE  160		//    


int main()
{
// 
	int n,i;
	struct{int size; int *ref;}test[COUNT]; 
//------------------------------------------------
	memset(test,0,sizeof(test));
	StubAllocator obj1(sizeof(int));
	for(n=0; n<COUNT*1000; n++) {
		i=rand()%COUNT;
		printf("[1]pass: %d(%02d%%)\r",n,(n+1)*100/(COUNT*1000));
		if(test[i].ref) {
			if(*test[i].ref==i) {printf("Internal error\n\a");getchar();return 1;}
			obj1.Free(test[i].ref),test[i].ref=0;
		}else{
			test[i].ref=(int*)obj1.New();
		}
	}
	putchar('\n');
//------------------------------------------------
	Allocator obj2;
	memset(test,0,sizeof(test));;
	for(n=0; n<COUNT*1000; n++) {
		i=rand()%COUNT;
		printf("[2]pass: %d(%02d%%)\r",n,(n+1)*100/(COUNT*1000));
		if(test[i].ref) {
			if(*test[i].ref==i) {printf("Internal error\n\a");getchar();return 1;}
			obj2.Free(test[i].ref,test[i].size),test[i].ref=0;
		}else{
			test[i].size=rand()%SIZE+1;
			test[i].ref=(int*)obj2.New(test[i].size);
		}
	}
	putchar('\n');
	for(n=i=0; i<COUNT; i++) if(test[i].ref) n+=test[i].size;
	obj2.Stat();
	printf("use space: %d\n",n);
	return 0;
}
